import itertools
from typing import Any, Dict, Iterable, List, Optional, Set, TextIO, Tuple, Union

from ordered_set import OrderedSet

from ensure_sops import MissingSOPSMeta, UnencryptedItemsError, UnknownFormatError
from ensure_sops.enums import Formats, Methods


def _check_encryption(
    container: Union[List, Dict], prefix: str = ""
) -> Tuple[Set[str], Set[str]]:
    successful_keys = OrderedSet([])
    failed_keys = OrderedSet([])

    iterable: Iterable[Tuple[Any, Any]]
    if isinstance(container, dict):
        iterable = container.items()
    else:
        iterable = enumerate(container)

    for key, value in iterable:
        if (
            isinstance(value, str) and value.startswith("ENC[") and value.endswith("]")
        ) or value in ("", None):
            successful_keys.add(f"{prefix}{key}")
        elif isinstance(value, (dict, list)):
            successful_subkeys, failed_subkeys = _check_encryption(
                value, prefix=f"{prefix}{key}."
            )
            successful_keys.update(successful_subkeys)
            failed_keys.update(failed_subkeys)
        else:
            failed_keys.add(f"{prefix}{key}")
    return successful_keys, failed_keys


class SopsValidator:
    _file_cache: Optional[str]
    fmt: str

    def __init__(self, stream: TextIO, filename: str, method: Methods = Methods.strict):
        self.stream = stream
        self.filename = filename
        self.method = method
        self._parsers = self._determine_parsers(self.filename)

    def _determine_parsers(self, filename: str) -> List[Formats]:
        suffixes = filename.split(".")[1:]

        hinted_fmt = None
        for fmt, suffix in itertools.product(Formats, suffixes):
            if suffix in fmt.value.extensions:
                hinted_fmt = fmt
                break

        if self.method == Methods.strict:
            if hinted_fmt is not None:
                return [hinted_fmt]
            else:
                return [Formats.json, Formats.yaml]  # Binary files use json by default

        reordered_formats = list(Formats)
        if hinted_fmt:
            reordered_formats.insert(
                0, reordered_formats.pop(reordered_formats.index(hinted_fmt))
            )
        return reordered_formats

    def _cache_file(self) -> None:
        position = self.stream.tell()
        file_content = self.stream.read()
        self._file_cache = file_content
        self.stream.seek(position)

    def check_encryption(self, fmt: Formats, values: Dict[str, Any]) -> None:
        if fmt == Formats.bin:
            raise UnknownFormatError([fmt_class for fmt_class in self._parsers])
        else:
            has_sops_keys, user_defined_values = fmt.value.filter_values(values)
            successful_keys, failed_keys = _check_encryption(user_defined_values)
            if failed_keys:
                raise UnencryptedItemsError(failed_keys)
            elif not has_sops_keys:
                raise MissingSOPSMeta(list(values.keys()))

    def parse(self) -> Tuple[Formats, Dict[str, Any]]:
        self._cache_file()
        for fmt in self._parsers:
            is_success, values = fmt.value.parse(self._file_cache)
            if is_success:
                return fmt, values

        self._file_cache = None  # Cleanup Memory
        return Formats.bin, {}



import configparser
import io
import json
import logging
import re
from typing import Any, Dict, List, Optional, Tuple

import dotenv
from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError

yaml = YAML(typ="safe")

logger = logging.getLogger("dotenv.main")
logger.setLevel(logging.ERROR)


class Format:
    name: str
    extensions: List[str]
    ignore_pattern: re.Pattern

    def parse(
        self, raw_string: str
    ) -> Tuple[bool, Optional[Dict[str, Any]]]:  # pragma: no cover
        return NotImplemented

    def filter_values(self, values: Dict[str, Any]) -> Tuple[bool, Dict[str, Any]]:
        has_sops_keys = False
        user_defined_values = {}
        ignore_pattern: re.Pattern = self.ignore_pattern
        for key, value in values.items():
            if not ignore_pattern.match(key):
                user_defined_values[key] = value
            else:
                has_sops_keys = True
        return has_sops_keys, user_defined_values


class JsonFormat(Format):
    name: str = "json"
    extensions: List[str] = ["json"]
    ignore_pattern: re.Pattern = re.compile("^sops$")

    def parse(self, raw_string: str) -> Tuple[bool, Optional[Dict[str, Any]]]:
        try:
            values = json.loads(raw_string)
            if not isinstance(values, dict):
                return False, None
        except json.JSONDecodeError:
            return False, None
        else:
            return True, values


class YamlFormat(Format):
    name: str = "yaml"
    extensions: List[str] = ["yaml", "yml"]
    ignore_pattern: re.Pattern = re.compile("^sops$")

    def parse(self, raw_string: str) -> Tuple[bool, Optional[Dict[str, Any]]]:
        try:
            values = yaml.load(raw_string)
            if not isinstance(values, dict):
                return False, None
        except ParserError:
            return False, None
        else:
            return True, values


class IniFormat(Format):
    name: str = "ini"
    extensions: List[str] = ["ini", "cfg", "conf", "config"]
    ignore_pattern: re.Pattern = re.compile("^sops$")

    def ini_to_dict(self, parser):
        results = {}
        for section in parser.sections():
            results[section] = {}
            for key, value in parser.items(section):
                results[section][key] = None if value == "" else value
        return results

    def parse(self, raw_string: str) -> Tuple[bool, Optional[Dict[str, Any]]]:
        parser = configparser.ConfigParser()
        try:
            parser.read_string(raw_string)
            values = self.ini_to_dict(parser)
            if not values:
                return False, None
        except configparser.ParsingError:
            return False, None
        else:
            return True, values


class EnvFormat(Format):
    name: str = "env"
    extensions: List[str] = ["env"]
    ignore_pattern: re.Pattern = re.compile("^sops_.+")

    def parse(self, raw_string: str) -> Tuple[bool, Optional[Dict[str, Any]]]:
        values = dotenv.dotenv_values(stream=io.StringIO(raw_string))
        cleaned_values = {
            key: value for key, value in values.items() if value not in ("", None)
        }
        if not bool(cleaned_values):
            return False, None
        else:
            return True, cleaned_values


class BinFormat(Format):
    name: str = "bin"
    extensions: List[str] = []
    ignore_pattern: re.Pattern = re.compile("^$")

    def parse(self, raw_string: str) -> Tuple[bool, Optional[Dict[str, Any]]]:
        return False, None



import enum

from .formats import BinFormat, EnvFormat, IniFormat, JsonFormat, YamlFormat


class Formats(enum.Enum):
    json = JsonFormat()
    ini = IniFormat()
    yaml = YamlFormat()
    env = EnvFormat()
    bin = BinFormat()


class Methods(enum.Enum):
    """
    strict: Try either the one matching with file extension or json
    bruteforce: Try all, starting from the one which matches with file extensions
    """

    strict = "strict"
    bruteforce = "bruteforce"

    def __str__(self):
        return self.name



import pathlib
import sys
from argparse import ArgumentParser
from typing import Dict, List, Optional

from .exceptions import ValidationError
from .validator import Methods, SopsValidator


def _get_parser() -> ArgumentParser:
    argparser = ArgumentParser()
    argparser.add_argument("files", type=pathlib.Path, nargs="+")
    argparser.add_argument(
        "--method", type=Methods, choices=list(Methods), default=Methods.strict
    )
    return argparser


def _validate_files(
    paths: List[pathlib.Path], method: Methods
) -> Dict[pathlib.Path, str]:
    failed_files = {}
    for path in paths:
        with path.open() as stream:
            validator = SopsValidator(stream, path.name, method=method)
            fmt, values = validator.parse()
            try:
                validator.check_encryption(fmt, values)
            except ValidationError as e:
                failed_files[path] = str(e)
    return failed_files


def main(args: Optional[List[str]] = None) -> int:
    parser = _get_parser()
    parsed_args = parser.parse_args(args)

    errors = _validate_files(parsed_args.files, parsed_args.method)

    if errors:
        for path, error in errors.items():
            print(f"{path} - {error}", file=sys.stderr)
        return 1
    return 0



from typing import Any, Dict, List, Set

from ensure_sops import Formats


class ValidationError(Exception):
    message: str
    data: Dict[str, Any]

    def __str__(self):
        return self.message.format(**self.data)


class UnknownFormatError(ValidationError):
    message = (
        "File format cannot be determined, here's the list of formats tried: "
        "{tried_formats}"
    )

    def __init__(self, tried_formats: List[Formats]):
        self.data = {"tried_formats": [fmt.name for fmt in tried_formats]}


class UnencryptedItemsError(ValidationError):
    message = "Some items in this file are not encrypted, here's the list: " "{items}"

    def __init__(self, items: Set[str]):
        self.data = {"items": items}


class MissingSOPSMeta(ValidationError):
    message = (
        "SOPS metadata is missing, here's all the top level keys in the file: "
        "{found_keys}"
    )

    def __init__(self, found_keys: List[str]):
        self.data = {"found_keys": found_keys}



from .enums import Formats, Methods
from .exceptions import (
    MissingSOPSMeta,
    UnencryptedItemsError,
    UnknownFormatError,
    ValidationError,
)
from .formats import BinFormat, EnvFormat, Format, IniFormat, JsonFormat, YamlFormat
from .main import main
from .validator import SopsValidator

__all__ = [
    "Formats",
    "Methods",
    "SopsValidator",
    "main",
    "ValidationError",
    "UnencryptedItemsError",
    "UnknownFormatError",
    "MissingSOPSMeta",
    "Format",
    "JsonFormat",
    "YamlFormat",
    "IniFormat",
    "EnvFormat",
    "BinFormat",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ensure_sops']

package_data = \
{'': ['*']}

install_requires = \
['ordered-set>=4.1.0,<5.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'ruamel.yaml>=0.17.21,<0.18.0']

entry_points = \
{'console_scripts': ['ensure_sops = ensure_sops.main:main']}

setup_kwargs = {
    'name': 'ensure-sops',
    'version': '0.1.2',
    'description': 'Check if your files are encrypted with Mozilla SOPS or not. Can act as pre-commit hook as well.',
    'long_description': None,
    'author': 'Engineering Team at Tacto Technology GmbH.',
    'author_email': 'engineering@tacto.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
